#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: container.tcl,v 1.5 2000/10/15 17:19:20 artur Exp $
#
# Information:
# Modul (benutzt von widgetlib.tcl)
# Eine Biblothek mit Container und Iteratoren in XOTcl
# DoubleList ist entworfen an die Anlehnung an c++ STL
# Allein wegen der Listen hat allerdings den zweifallhaften Sinn
# weil Tcl eigene Listen hat.
# Vielmehr wird hier aus der Benutzung der Iteratoren, Visitors 
# Classen (foreach_elem) oder Compositemglichkeiten gebraucht gemacht.

Class DobleListNode -parameter {
    {_next NIL}
    {_prev NIL}
}
Class DLIterator -parameter {
    node
}
DLIterator instproc get {} {
    return [set [self]::node]
}
# Entsprich i++ bei STL C++
DLIterator instproc next_elem {} {
    [self] instvar node
    if {$node=="NIL"} {
	return NIL
    } else {
	set ret $node
	set node [set ${node}::dln::_next]
	return $ret
    }
}
# Entsprinch i-- bei STL C++
DLIterator instproc prev_elem {} {
    [self] instvar node
    if {$node=="NIL"} {
	return NIL
    } else {
	set ret $node
	set node [set ${node}::dln::_prev]
	return $ret
    }    
}

Class DoubleList -parameter {
    {_head NIL}
    {_tail NIL}
}
# Die Elemente des Containers mssen in im existieren
# also fr Container db muss das Element db::element heien

DoubleList instproc push_back eclass {
    [self] instvar _tail _head
    if {[[self] empty]} {
	DobleListNode ${eclass}::dln
	set _head $eclass
	set _tail $eclass    
    } else {
	DobleListNode ${eclass}::dln -_prev $_tail 
	set ${_tail}::dln::_next $eclass
	set _tail $eclass
    }
}
DoubleList instproc push_head eclass {
    [self] instvar _tail _head
    if {[[self] empty]} {
	DobleListNode ${eclass}::dln
	set _tail $eclass
	set _head $eclass
    } else {
	DobleListNode ${eclass}::dln -_next $_head
	set ${_head}::dln::_prev $eclass
	set _head $eclass
    }
}
DoubleList instproc empty {} {
    [self] instvar _tail _head
    if {$_head=="NIL" && $_tail=="NIL"} {
	return 1
    } else {
	return 0
    }
}
DoubleList instproc begin_iterator {name} {
    [self] instvar _head
    return [DLIterator $name -node $_head]
}
DoubleList instproc getelem_list {} {
    [self] instvar _head
    DLIterator [self]::it -node $_head
    while {[[self]::it get]!="NIL"} {
	lappend ret [[self]::it next_elem]
    }
    [self]::it destroy
    return $ret
}
DoubleList instproc end_list {} {
    return NIL
}
DoubleList instproc foreach_elem {class method {classfilter *}} {
    [self] instvar _head
    DLIterator [self]::it -node $_head
    while {[[self]::it get]!="NIL"} {
	if {$classfilter=="*"} {
	    $class $method [[self]::it next_elem]
	} else {
	    set elem [[self]::it next_elem]
	    if {[string match $classfilter [$elem info class]]} {
		$class $method [[self]::it next_elem]
	    }
	}
    }
    [self]::it destroy
}
DoubleList instproc foreach_elem_deep {class method {classfilter *}} {
    [self] instvar _head
    DLIterator [self]::it -node $_head
    while {[[self]::it get]!="NIL"} {
	set act_elem [[self]::it next_elem]
	if {$classfilter=="*"} {
	    $class $method $act_elem
	    if 
	} else {
	    set elem [[self]::it next_elem]
	    if {[string match $classfilter [$elem info class]]} {
		$class $method $act_elem
	    }
	}
	# Wenn actualles Alement hat DoubleList dann rekursiv aufrufen
	if {[lsearch [[$act_elem info class] info heritage] DoubleList]!=-1} {
	    $act_elem foreach_elem_deep $clss $method $classfilter
	}
    }
    [self]::it destroy
}
