####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id
#
# Information:
# Modul (benutzt von FormEngine)
# Wird mit einem Formular verbunden
# benutzt um SQL Abfrage als Prozeduren auf Tupel zu Kapseln
# DBProxy ist ein Vermitler zwischen SQL Verbindung und Engine
# Er ist spezielisiert Tabelen abzufragen und Ergebnisse nach Index zu Liefern
# auch Tupel Anzahl mu er speichern
# Er wird bentigt um vielleicht Ergebnisse zu puffern
# Die unterschiedliche Arten von Abfragen zu kapseln
# oder andere als SQl Interfaces zu realisieren
#
# Hier kann auch die Zentrale Stelle sein um neue oder modefizierte 
# Tupel zu merken
# sqlconnect - name des SQL Objekt 
# table - Tabellen Name
# attrlist - Attributten Liste
# primarykey - Prim Key benutzt fr Identifikation beim Modefizieren oder Lschen
# refvalue und refattribute - wird zu WHERE refattribute="$refvalue" transferiert
Class DBProxy -parameter {sqlconnect table attrlist primarykey refvalue {refattribute {}}}
DBProxy instproc init args {
    [self] instvar attrlist primarykey  qdata refattribute qfilter
    if {![lcontain $attrlist $refattribute] && $refattribute!=""} {
	lappend attrlist $refattribute
    }
    set notdata [minusMenge $primarykey $attrlist]
    if {[lempty $notdata]} {
	set qdata $attrlist
    } else {
	set qdata [concat $attrlist $notdata]
    }
    set qfilter {}
}
DBProxy instproc reload {} {
    [self] instvar sqlconnect table attrlist qdata primarykey qfilter query lastfetch refattribute refvalue orderby
    if {[info exist [self]::query]} {
	$query destroy
    }
    set sqlstatment "SELECT [join $qdata ,] FROM $table"
    set where ""
    if {$refattribute!=""} {
	append where " WHERE $refattribute=$refvalue"
    }
    if {$qfilter!=""} {
	if {$where==""} {
	    append where " WHERE $qfilter"
	} else {
	    append where " AND ( $qfilter )"
	}
    }
    append sqlstatment $where
    if {[info exist orderby] && $orderby!=""} {
	append sqlstatment " ORDER BY $orderby"
    }
    #puts "QUERY $sqlstatment"
    set query [$sqlconnect query $sqlstatment]
    set lastfetch -1
    return [$query rows]
}
# Nimmt ein Tupel als tclx keyedlist
# lastfetch ist ein ID des letzten genommenen Tuppels
# es wird immer die Identifikation des letzten Tuppel gespeichert
# als keyedlist tupelid
# um in nacher modifizieren oder lschen zu knnen
DBProxy instproc gettupel {id} {
    [self] instvar query lastfetch attrlist qdata tupelid primarykey
    if {$id==[expr $lastfetch+1]} {
	# Einfach der nschtr
	set row [$query fetch]
	incr lastfetch
    } else {
	#Pointer setzten
	$query seek $id
	set lastfetch $id
	set row [$query fetch]
    }
    #puts "GET $row"
    set tupel {}
    set index 0
    foreach column $attrlist {
	keylset tupel $column [lindex $row $index]
	incr index
    }
    foreach keyattr $primarykey {
	keylset tupelid $keyattr [lindex $row [lsearch $qdata $keyattr]]
    }
    return $tupel
}
# Es werden Zeichenketten als SQL Strings maskiert
# Ersetzung von " und \
# und Umschlieen mit "............"
# NULL wird nicht maskiert
DBProxy instproc maskstring {string} {
    if {$string=="NULL"} {return NULL}
    regsub -all {\\} $string {\\\\} string
    regsub -all \" $string {\"} string
    return "'$string'"  
}
# Baut die richtige WHERE Klausel aus tupelid (interne Benutzung)
DBProxy instproc buildwhere {} {
    [self] instvar tupelid refattribute refvalue
    set where {}
    foreach attr [keylkeys tupelid] {
	if {$where!=""} { append where " AND " }
	if {[set val  [keylget tupelid $attr]]!="NULL"} {
	    append where "$attr=[[self] maskstring $val]"
	} else {
	    append where "$attr IS NULL"
	}
    }
    # Bei Vererbung kann ein refattribute zum prim gehren
    # also keine verdopplung
    if {$refattribute!="" && ![lcontain [keylkeys tupelid] $refattribute]} {
	append where " AND $refattribute=$refvalue"
    }
    return $where
}
# Baut einen Filter (oder Suchfunktion) ber die Tupel
# filterattr - ist eine Keydlist (TclX) ber die Attributen
# boolmode - kann "and" oder "or" sein (komplexe Abfragen so nicht mglich)
# filtermatch - kann "exact" "glob" oder "regexp" sein die Abfrage wird 
#               gebildet mit = Like und =~
# Die einzelnen Attributfilter knnen modifikatoren haben
# = > < >= <= ! (not)  ~ (regexp) * (glob)
# \ maskiert den Modifiekator
DBProxy instproc setFilter {filterattr boolmode filtermatch} {
    [self] instvar qfilter
    set qfilter {}
    foreach attr [keylkeys filterattr] {
	if {$qfilter!=""} {
	    if {$boolmode=="and"} {
		append qfilter " AND "
	    } else {
		append qfilter " OR "
	    }
	}
	set val [keylget filterattr $attr]
	set mod [string range $val 0 0]
	set op {}
	set noval 0
	switch -- $mod {
	    = -
	    > -
	    < {
		set val [string range $val 1 end]
		set op $mod
	    }
	    * {
		set val [string range $val 1 end]
		set op LIKE
	    }
	    ~ {
		set val [string range $val 1 end]
		set op REGEXP
	    }
	    ! {
		switch [string range $val 1 1] {
		    = {
			set val [string range $val 2 end]
			set op !=
		    }
		    * {
			set val [string range $val 2 end]
			set op "NOT LIKE"
		    }
		    ~ {
			set val [string range $val 2 end]
			set op "NOT REGEXP"
		    }
		    default {
			set val [string range $val 1 end]
			set op !=
		    }
		}
	    }
	    \\ {
		set val [string range $val 1 end]
	    }
	    default {
		switch $val {
		    "IS NULL" {
			set op "IS NULL"
			set noval 1
		    }
		    "IS NOT NULL" {
			set op "IS NOT NULL"
			set noval 1
		    }
		}
	    }
	}
	#puts "OP $op mod=$mod"
	if {$op==""} {
	    switch $filtermatch {
		glob {
		    set op LIKE
		}
		regexp {
		    set op REGEXP
		}
		exact -
		default {
		    set op =
		}
	    }
	}
	if {$noval} {
	    append qfilter "$attr $op"
	} else {
	    append qfilter "$attr $op [[self] maskstring $val]"
	}
    }
    #puts "setFilter $filterattr $qfilter"
}
# Sortierung 
# sortlist= {xattr1 xattr2 xattr3}
# xattr1 = {attr1 desc} | attr1
# Es wird zu liste by SQL Abfrage
# ... ORDER BY xattr1, xattr2, xattr3
DBProxy instproc setOrder {sortlist} {
    [self] instvar orderby
    set orderby [join $sortlist ,]
}
# Fht die nderungen an dem letzten gelesenen Tupel
DBProxy instproc update {changes} {
    [self] instvar sqlconnect table attrlist
    set updateset {}
    foreach attr [keylkeys changes] {
	if {![lcontain $attrlist $attr]} continue
	if {$updateset!=""} { append updateset "," }
	append updateset "$attr=[[self] maskstring [keylget changes $attr]]"
    }
    $sqlconnect execute "UPDATE $table SET $updateset WHERE [[self] buildwhere]"
}
DBProxy instproc delete {} {
    [self] instvar sqlconnect table
    $sqlconnect execute "DELETE FROM $table WHERE [[self] buildwhere]"
}
# Gibt zurck den primrschlssel des neuen Tupel fall aus einen Attribut
# sonst NOPRIM
DBProxy instproc new {values} {
    [self] instvar sqlconnect table refattribute refvalue primarykey
    set newvalues {}
    foreach attr [keylkeys values] {
	if {$newvalues!=""} { append newvalues ","; append collist "," }
	append newvalues [maskstring [keylget values $attr]]
	append collist $attr
    }
    if {$refattribute!=""} {
	append newvalues ",$refvalue"
	append collist ",$refattribute"
    }
    $sqlconnect execute "INSERT INTO $table ($collist) VALUES ($newvalues)"
    if {[llength $primarykey]!=1} {
	return NOPRIM
    } else {
	if {[lcontain [keylkeys values] $primarykey]} {
	    return [keylget values $primarykey]
	} else {
	    return [$sqlconnect getinsertid]
	}
    }
}
DBProxy instproc destroy {} {
    [self] instvar  query 
    if {[info exist query]} {
	$query destroy
    }
    next
}
# Diese Klasse wird fr Formulare verwendet die 
# ein ObjektLink zwischen n:m Bezihung reprsentieren
# Die Abfrage etwa
# select table.attrlist1,table.attrlist2,... from tabel,reftable
#        where reftable.refattribute2=table.primarykey and
#              reftable.refattribute="refvalue"
Class DBProxyRT -superclass {DBProxy} -parameter {
    sqlconnect table attrlist primarykey 
    refvalue refattribute reftable refattribute2
}
DBProxyRT instproc init args {
    [self] instvar attrlist primarykey  qdata refattribute qfilter
    set notdata [minusMenge $primarykey $attrlist]
    if {[lempty $notdata]} {
	set qdata $attrlist
    } else {
	set qdata [concat $attrlist $notdata]
    }
    set qfilter {}
}
DBProxyRT instproc reload {} {
    [self] instvar sqlconnect table attrlist qdata primarykey qfilter query lastfetch \
	refattribute refattribute2 refvalue reftable
    if {[info exist [self]::query]} {
	$query destroy
    }
    set sqlstatment "SELECT $table.[join $qdata ,$table.] FROM $table,$reftable"
    set where ""
    if {$refattribute!=""} {
	append where " WHERE $reftable.$refattribute=$refvalue AND $reftable.$refattribute2=$table.$primarykey"
    }
    if {$qfilter!=""} {
	if {$where==""} {
	    append where " WHERE $qfilter"
	} else {
	    append where " AND ( $qfilter )"
	}
    }
    append sqlstatment $where
    #puts "QUERY $sqlstatment"
    set query [$sqlconnect query $sqlstatment]
    set lastfetch -1
    return [$query rows]
}
# Nimmt ein Tupel als tclx keyedlist
# lastfetch ist ein ID des letzten genommenen Tuppels
# es wird immer die Identifikation des letzten Tuppel gespeichert
# als keyedlist tupelid
# um in nacher modifizieren oder lschen zu knnen
DBProxyRT instproc gettupel {id} {
    [self] instvar query lastfetch attrlist qdata tupelid primarykey
    if {$id==[expr $lastfetch+1]} {
	# Einfach der nschtr
	set row [$query fetch]
	incr lastfetch
    } else {
	#Pointer setzten
	$query seek $id
	set lastfetch $id
	set row [$query fetch]
    }
    #puts "GET $row"
    set tupel {}
    set index 0
    foreach column $attrlist {
	keylset tupel $column [lindex $row $index]
	incr index
    }
    foreach keyattr $primarykey {
	keylset tupelid $keyattr [lindex $row [lsearch $qdata $keyattr]]
    }
    return $tupel
}
# Baut die richtige WHERE Klausel aus tupelid (interne Benutzung)
DBProxyRT instproc buildwhere {} {
    [self] instvar tupelid
    set where {}
    foreach attr [keylkeys tupelid] {
	if {$where!=""} { append where " AND " }
	if {[set val  [keylget tupelid $attr]]!="NULL"} {
	    append where "$attr=[[self] maskstring $val]"
	} else {
	    append where "$attr IS NULL"
	}
    }
    return $where
}
DBProxyRT instproc delete {} {
    bgerror "Forbiden Operation on this kind of proxy (objectlink)"
}
# Gibt zurck den primrschlssel des neuen Tupel fall aus einen Attribut
# sonst NOPRIM
DBProxyRT instproc new {values} {
    bgerror "Forbiden Operation on this kind of proxy (objectlink)"
}


